#include <QDomDocument>
#include <QDomElement>

#include <QFileDialog>

#include <QMessageBox>

#include "mnkPasswordKeeperMainWindow.H"
#include "mnkPasswordMakerDialog.H"

mnkPasswordKeeperMainWindow::mnkPasswordKeeperMainWindow(QMainWindow *parent)
  : QMainWindow(parent)
{
  setupUi(this);
  addDockWidget(Qt::LeftDockWidgetArea, dockWidget);

  fileName = QString(DEFAULTFILE);
  currentPassword = NULL;
  currentItem = NULL;
  generatorDialog = NULL;

  connect(removeButton, SIGNAL(clicked(void)),
          this, SLOT(removeButtonPushed(void)));
  connect(newButton, SIGNAL(clicked(void)),
          this, SLOT(newButtonPushed(void)));

  connect(filterEdit, SIGNAL(textChanged(const QString&)),
          this, SLOT(filterEdited(const QString&)));

  connect(actionOpen, SIGNAL(triggered(void)),
          this, SLOT(fileOpen(void)));
  connect(actionSave, SIGNAL(triggered(void)),
          this, SLOT(fileSave(void)));
  connect(actionSave_As, SIGNAL(triggered(void)),
          this, SLOT(fileSaveAs(void)));
  connect(actionClose, SIGNAL(triggered(void)),
          this, SLOT(fileClose(void)));
  connect(actionExit, SIGNAL(triggered(void)),
          this, SLOT(fileExit(void)));

  connect(actionDock, SIGNAL(triggered(void)),
          this, SLOT(dockShow(void)));
  connect(actionAbout, SIGNAL(triggered(void)),
          this, SLOT(aboutShow(void)));
  connect(actionMaster_Password, SIGNAL(triggered(void)),
          this, SLOT(showMasterDialog(void)));
  connect(actionGenerator, SIGNAL(triggered(void)),
          this, SLOT(showGenerator(void)));

  connect(siteEdit, SIGNAL(textEdited(const QString&)),
          this, SLOT(applyTextChanged(const QString&)));
  connect(usernameEdit, SIGNAL(textEdited(const QString&)),
          this, SLOT(applyTextChanged(const QString&)));
  connect(passwordEdit, SIGNAL(textEdited(const QString&)),
          this, SLOT(applyTextChanged(const QString&)));
  connect(commentTextEdit, SIGNAL(textChanged()),
          this, SLOT(applyButtonPushed()));

//  siteEdit / usernameEdit / passwordEdit / commentTextEdit

  connect(listWidget,
          SIGNAL(currentItemChanged(QListWidgetItem*,QListWidgetItem*)),
          this,
          SLOT(itemChanged(QListWidgetItem*, QListWidgetItem*)));

  actionFilterFocus = new QAction(this);
  actionFilterFocus->setShortcut(tr("Ctrl+F"));
  connect(actionFilterFocus, SIGNAL(triggered(void)),
          this, SLOT(filterFocus(void)));

  this->addAction(actionFilterFocus);

  fileName = QString();

  myCryptInterface = new mnkCrypt();
  myCryptInterface->mnkSetKey(QString("this is my key").toAscii());

  masterPassword = QString();
  showMasterDialog();
  if(masterPassword == QString()) {
    QMessageBox::critical(this, tr("Error"), tr("No master password defined."));
    exit(-1);
  }

  if(QFile::exists(DEFAULTFILE) == true) {
    if(openReadFile(DEFAULTFILE) == true) 
      this->fileName = DEFAULTFILE;
    reloadList();
  }
  isModified = false;
}

mnkPasswordKeeperMainWindow::~mnkPasswordKeeperMainWindow(void)
{
  return ;
}

void mnkPasswordKeeperMainWindow::removeButtonPushed(void)
{
  mnkPassword *myPassword;
  QListWidgetItem *myItem = listWidget->currentItem();

  if(myItem == NULL)
    return;

  myPassword = qVariantValue<mnkPassword*>(myItem->data(Qt::UserRole));

  int ret = QMessageBox::warning(this,
                                 tr("Password deletion"),
                                 tr("Do you want to delete '%1' ?")
                                   .arg(myPassword->site),
                                 QMessageBox::Yes | QMessageBox::No);

  if(ret == QMessageBox::Yes) {
    currentPassword = NULL;
    statusbar->showMessage(tr("Record removed: %1").arg(myPassword->site), STM);

    passwordList.removeAll(myPassword);
    isModified = true;
    reloadList();
  }
}

void mnkPasswordKeeperMainWindow::newButtonPushed(void)
{
  QListWidgetItem *myItem;
  mnkPassword *myPassword = new mnkPassword();

  currentPassword = NULL;
  currentItem = NULL;

  listWidget->clearSelection();

  siteEdit->setText(QString("<< New record >>"));
  usernameEdit->setText(QString());
  passwordEdit->setText(QString());
  commentTextEdit->setText(QString());

  myPassword->site = siteEdit->text();

  myItem = new QListWidgetItem();
  myItem->setData(Qt::UserRole, qVariantFromValue(myPassword));
  myItem->setText(QString("%1 (%2)")
                     .arg(myPassword->site, myPassword->username));

  listWidget->addItem(myItem);
  passwordList.append(myPassword);
  currentPassword = myPassword;

  listWidget->setCurrentItem(myItem);

  isModified = true;

  siteEdit->setFocus();
  siteEdit->selectAll();
}

void mnkPasswordKeeperMainWindow::applyButtonPushed(void)
{
#ifdef DEBUG
  printf("[CALL: applyButtonPushed]\n");
#endif

  mnkPassword *myPassword;

  if(currentPassword == NULL
  || currentItem == NULL)
    return;

  myPassword = currentPassword;

  if(currentItem != NULL) {
    currentItem->setText(QString("%1 (%2)")
                    .arg(siteEdit->text(), usernameEdit->text()));
  }

  myPassword->site = siteEdit->text();
  myPassword->username = usernameEdit->text();
  myPassword->password = passwordEdit->text();
  myPassword->comment = commentTextEdit->toPlainText();

  statusbar->showMessage(tr("Record modified: %1").arg(myPassword->site), STM);

  isModified = true;
}

void mnkPasswordKeeperMainWindow::applyTextChanged(const QString &edt)
{
  applyButtonPushed();
}

void mnkPasswordKeeperMainWindow::fileOpen(const QString &path)
{
  QString fileName = path;

  if(isModified == true) {
    bool ret = fileClose();
    if(ret == false)
      return;
  }

  if(true == fileName.isNull())
    fileName = QFileDialog::getOpenFileName(this,
                                            tr("Open File"),
                                            "",
                                            "Password Files (*.xmls)");

  if(false == fileName.isEmpty()) {
    this->openReadFile(fileName);
    reloadList();
    this->fileName = fileName;
    isModified = false;
  }
}

void mnkPasswordKeeperMainWindow::fileSave(void)
{
  return(this->fileSaveAs(this->fileName));
}

void mnkPasswordKeeperMainWindow::fileSaveAs(const QString &path)
{
  QString lfile = path;

  QString tmpStr = QString("<pass>\n");
  
  for(int i = 0; i < passwordList.count(); i ++) {
    mnkPassword *m1 = passwordList.at(i);
    tmpStr+= QString("  <password>\n");
    tmpStr+= QString("    <site>%1</site>\n").arg(ENCODE(m1->site));
    tmpStr+= QString("    <username>%1</username>\n").arg(ENCODE(m1->username));
    tmpStr+= QString("    <password>%1</password>\n").arg(ENCODE(m1->password));
    tmpStr+= QString("    <comment>%1</comment>\n").arg(ENCODE(m1->comment));
    tmpStr+= QString("  </password>\n");
  }
  
  tmpStr += QString("</pass>\n\n");

  if(lfile.isEmpty() == true) {
    lfile = QFileDialog::getSaveFileName(this,
                                         tr("Save File"),
                                         "", 
                                         "Password Files (*.xmls)");
  }

  if(lfile.isEmpty() == true) {
    statusbar->showMessage(tr("Error while saving file: No file name"));
    return;
  }

  if(lfile.contains(".xmls") == false)
    lfile += QString(".xmls");
  
  QFile db(lfile);
  bool ret = db.open(QIODevice::WriteOnly);
  if(ret == false) {
    QMessageBox::warning(this, tr("Can't save file"),
                        db.errorString());
  } else {
    int ret = db.write(tmpStr.toAscii());
    if(ret == tmpStr.toAscii().size()) {
      this->fileName = lfile;
      statusbar->showMessage(tr("File %1 saved.").arg(lfile), STM);
      isModified = false;
    } else {
      statusbar->showMessage(tr("Error while saving file."), STM); 
    }
    db.close();
  }
}

bool mnkPasswordKeeperMainWindow::fileClose(void)
{
  bool isClosed = true;

  if(isModified == true) {
    int ret = QMessageBox::warning(this,
                tr("Changed not saved"),
                tr("Do you want to save changes ?"),
                QMessageBox::Yes 
                | QMessageBox::No
                | QMessageBox::Cancel);
    if(ret == QMessageBox::Yes) {
      fileSave();
      isClosed = true;
    } else if(ret == QMessageBox::No) {
      isClosed = true;
    } else
      isClosed = false;
  }

  if(isClosed == true) {
    /* Close */
    listWidget->clearSelection();

    siteEdit->setText(QString());
    usernameEdit->setText(QString());
    passwordEdit->setText(QString());
    commentTextEdit->blockSignals(true);
    commentTextEdit->setText(QString());

    listWidget->clear();
    commentTextEdit->blockSignals(false);

    /* clear pwd list */
    while(passwordList.count() > 0)
      delete(passwordList.takeAt(0));

    this->fileName = QString();
  } else
    return(false);

  return(true);
}

void mnkPasswordKeeperMainWindow::fileExit(void)
{
  if(true == fileClose())
    QApplication::exit(0);
}

void mnkPasswordKeeperMainWindow::itemChanged(QListWidgetItem*current, 
                                              QListWidgetItem*previous)
{
  mnkPassword *myPassword;

#ifdef DEBUG
  printf("[CALL: itemChanged]\n");
#endif

  if(current == NULL) {
    currentPassword = NULL;
    currentItem = NULL;

    listWidget->clearSelection();

    siteEdit->setText(QString());
    usernameEdit->setText(QString());
    passwordEdit->setText(QString());
    commentTextEdit->setText(QString());

    return;
  }

  myPassword = qVariantValue<mnkPassword*>(current->data(Qt::UserRole));

  currentPassword = myPassword;
  currentItem = current;

  siteEdit->setText(myPassword->site);
  usernameEdit->setText(myPassword->username);
  passwordEdit->setText(myPassword->password);
  commentTextEdit->blockSignals(true);
  commentTextEdit->setPlainText(myPassword->comment);
  commentTextEdit->blockSignals(false);

  /* Change status */

  return;
}

void mnkPasswordKeeperMainWindow::reloadList(void)
{
  QListWidgetItem *myItem, *firstItem = NULL;
  mnkPassword *myPassword;

  listWidget->clear();

  for(int i = 0; i < passwordList.count(); i ++) {
    myPassword = passwordList.at(i);
    if(filterEdit->text().isEmpty() == false) {
      QString txt = filterEdit->text();
      if(myPassword->site.contains(txt) == false
      && myPassword->username.contains(txt) == false
      && myPassword->password.contains(txt) == false
      && myPassword->comment.contains(txt) == false)
        continue;
    }

    myItem = new QListWidgetItem();
    myItem->setData(Qt::UserRole, qVariantFromValue(myPassword));
    myItem->setText(QString("%1 (%2)")
                    .arg(myPassword->site, myPassword->username));

    listWidget->addItem(myItem);
    if(firstItem == NULL)
      firstItem = myItem;
  }
  
  /* select 1st element */
  if(firstItem != NULL)
    listWidget->setCurrentItem(firstItem);

  listWidget->sortItems();
}

bool mnkPasswordKeeperMainWindow::openReadFile(QString fileName)
{
  QDomElement root, child;
  QDomDocument domDocument;
  QString errorStr;
  int errorLine, errorColumn;
  
  QFile file(fileName);
  
  QString site, password, username, comment;
  
  if(domDocument.setContent(&file, 
                            true,
			    &errorStr,
			    &errorLine,
			    &errorColumn) == false) {
    /* error */

    QString errTitle = QString("Error on line: %1 / column: %2")
                          .arg(QString::number(errorLine),
                               QString::number(errorColumn));

    QMessageBox::critical(this, errTitle, errorStr); 
    
    return(false);
  }
  
  root = domDocument.documentElement();
  
  child = root.firstChildElement("password");
  while(child.isNull() == false) {
    site = DECODE(child.firstChildElement("site").text());
    username = DECODE(child.firstChildElement("username").text());
    password = DECODE(child.firstChildElement("password").text());
    comment = DECODE(child.firstChildElement("comment").text());
    
    mnkPassword *newPass = new mnkPassword;
    newPass->site = site;
    newPass->username = username;
    newPass->password = password;
    newPass->comment = comment;
    
    passwordList.append(newPass);
    
    child = child.nextSiblingElement("password");
  }
  
  return(true);
}

void mnkPasswordKeeperMainWindow::filterEdited(const QString &txt)
{
  reloadList();
}

void mnkPasswordKeeperMainWindow::dockShow(void)
{
  if(dockWidget->isHidden() == true)
    dockWidget->show();
  else
    dockWidget->hide();
}

void mnkPasswordKeeperMainWindow::aboutShow(void)
{
  QString info = QString("Designed in Toulouse, France\n");
  info += QString("By Patrick MARIE <mycroft@monkeyz.eu>\n");
  info += QString("Check for updates on http://www.monkeyz.eu/ !\n");
  info += QString("Version: %1").arg(QString(VERSION));
  QMessageBox::information(this,
                    tr("About this software"),
                    info);
}

void mnkPasswordKeeperMainWindow::filterFocus(void)
{
  filterEdit->setFocus();
}

void mnkPasswordKeeperMainWindow::showMasterDialog(void)
{
  QDialog *myDialog;
  QWidget *buttons;
  QPushButton *okb, *notokb;
  QLabel *lab;
  QLineEdit *le;
  QHBoxLayout *hl;
  QVBoxLayout *vl;

  myDialog = new QDialog(this);
  myDialog->setWindowTitle(tr("Master password"));

  okb = new QPushButton(myDialog);
  okb->setText("Change");

  notokb = new QPushButton(myDialog);
  notokb->setText("Cancel");

  hl = new QHBoxLayout;
  hl->addWidget(okb);
  hl->addWidget(notokb);

  buttons = new QWidget(myDialog);
  buttons->setLayout(hl);

  lab = new QLabel(myDialog);
  lab->setText("Enter new Master password");

  le = new QLineEdit(myDialog);
  le->setEchoMode(QLineEdit::Password);

  vl = new QVBoxLayout;
  vl->addWidget(lab);
  vl->addWidget(le);
  vl->addWidget(buttons);
  
  myDialog->setLayout(vl);

  connect(okb, SIGNAL(clicked()),
          myDialog, SLOT(accept()));
  connect(notokb, SIGNAL(clicked()),
          myDialog, SLOT(reject()));
  
  le->setFocus();
  myDialog->exec();

  if(myDialog->result() == QDialog::Accepted) {
    masterPassword = le->text();
    myCryptInterface->mnkSetKey(masterPassword.toAscii());
  }
}

void mnkPasswordKeeperMainWindow::showGenerator(void)
{
  if(generatorDialog == NULL)
    generatorDialog = new mnkPasswordMakerDialog(this);

  if(generatorDialog->isHidden())
    generatorDialog->show();
  else
    generatorDialog->hide();
}
